////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#
include "config_1_2_3.h"
// threading library
#
include "pt_cornell_1_2_3.h"
// yup, the expander
#
include "port_expander_brl4.h"

////////////////////////////////////
// graphics libraries
// SPI channel 1 connections to TFT
#
include "tft_master.h"#
include "tft_gfx.h"
// need for rand function
#
include < stdlib.h >
    // need for sin function
    #include < math.h >
    ////////////////////////////////////

    // string buffer
    char buffer[60];

////////////////////////////////////
// DAC ISR
// A-channel, 1x, active
#
define DAC_config_chan_A 0b0011000000000000
// B-channel, 1x, active
# define DAC_config_chan_B 0b1011000000000000
volatile SpiChannel spiChn = SPI_CHANNEL2; // the SPI channel to use
// for 60 MHz PB clock use divide-by-3
volatile int spiClkDiv = 2; // 20 MHz DAC clock

//port b pulldown
#
define EnablePullDownB(bits) CNPUBCLR = bits;
CNPDBSET = bits;
//absolute function
#
define abs(a)((a < 0) ? (-a) : a)

// Microphone Data parameters
# define num_mic_channel 3# define recording_length 1 / 10# define sampling_rate 25000# define mic_data_size(sampling_rate * recording_length) //+correlate_window)
// Direction calculation parameters
# define speed 343# define sample_speed(speed / sampling_rate)

static short mic_bias[num_mic_channel] = {
    9,
    8,
    1
};
static short mic_data[num_mic_channel][mic_data_size];
static unsigned long correlate_data[num_mic_channel][mic_data_size >> 2];
static int peak_index[num_mic_channel];
static unsigned int peak_max[num_mic_channel];
static int correlate_shift;

//FLAGS
static int ready = 0;

// states for the push button
enum Button_State {
    RELEASED = 0, MAYBE_PRESSED = 1, PRESSED = 2, MAYBE_RELEASED = 3
};
static int ButtonState;
static int button;

static unsigned int index_0, index_1;

void __ISR(_TIMER_3_VECTOR, ipl2) Timer3Handler(void) {
    mT3ClearIntFlag();
    // === DAC Channel A =============
    // CS low to start transaction
    mPORTBClearBits(BIT_4); // start transaction
    // write to spi2
    WriteSPI2(DAC_config_chan_A | ((mic_data[1][index_0] + 512) & 0x3FF) << 2);
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
    // CS high
    mPORTBSetBits(BIT_4); // end transaction

    // === Channel B =============
    // CS low to start transaction
    mPORTBClearBits(BIT_4); // start transaction
    // write to spi2
    WriteSPI2(DAC_config_chan_B | (correlate_data[1][index_1] >> 16));
    index_0 = (index_0 + 1 == mic_data_size) ? 0 : index_0 + 1;
    index_1 = (index_1 + 1 == mic_data_size >> 1) ? 0 : index_1 + 1;
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
    // CS high
    mPORTBSetBits(BIT_4); // end transaction
}

// string buffer
char buffer[60];

// === thread structures ============================================
// thread control structs
static struct pt pt_debug, pt_button, pt_computation;
// system 1 second interval tick
int sys_time_seconds;

// compute the direction of the source sound
// return 0,1,2 to match mic channel, -1 for None
void compute_direction() {
    short direction = -1;
    //encode the positiveity three peak index.
    short val = (peak_index[0] > 0) << 2 | (peak_index[1] > 0) << 1 | (peak_index[2] > 0);
    //determind the direction based on the relative peak index
    //when the index is positive, the first mic is closer than the second mic channel
    if (val == 0b100 || val == 0b110) {
        direction = 0;
    } else if (val == 0b010 || val == 0b011) {
        direction = 1;
    } else if (val == 0b001 || val == 0b101) {
        direction = 2;
    }
    // draw a pointer on the TFT pointing to the direction of the source
    tft_fillRoundRect(10, 120, 200, 16, 1, ILI9340_BLACK);
    tft_setCursor(10, 120);
    tft_setTextColor(ILI9340_WHITE);
    tft_setTextSize(2);
    sprintf(buffer, "dir= %d", direction);
    tft_writeString(buffer);
    tft_fillCircle(120, 200, 36, ILI9340_BLACK);
    if (direction == 0) {
        tft_fillTriangle(120 - 10, 200,
            120, 200 - 10,
            120 - 25, 200 - 25, ILI9340_WHITE);
    } else if (direction == 1) {
        tft_fillTriangle(120 - 6, 200 + 9,
            120 + 6, 200 + 9,
            120, 200 + 30, ILI9340_WHITE);
    } else if (direction == 2) {
        tft_fillTriangle(120 + 10, 200,
            120, 200 - 10,
            120 + 25, 200 - 25, ILI9340_WHITE);
    }
    tft_fillCircle(120, 200, 10, ILI9340_GREEN);
}

// cross-correlate each pair of microphone recordings
void cross_correlate() {
    int channel, idx, shift;
    for (channel = 0; channel < num_mic_channel; channel++) {
        // shift the kernel -1/8 to 1/8 of mic_data_size
        for (shift = -mic_data_size >> 3; shift < (mic_data_size >> 3); shift++) {
            long tmp_sum = 0;
            // kernel size is 3/4 of mic_data_size
            for (idx = (mic_data_size >> 3); idx < (mic_data_size) - (mic_data_size >> 3); idx++) {
                int idx2 = (idx + shift);
                int correlate_channel = (channel + 1 > 2) ? 0 : channel + 1;
                tmp_sum += ((mic_data[channel][idx] + mic_bias[channel]) *
                    (mic_data[correlate_channel][idx2] + mic_bias[correlate_channel])) >> 1;
            }
            // Take absolute value, since the positivity doesn't matter. A negative max peak
            // should be identified since it just the wave shifted by 180 deg.
            correlate_data[channel][shift + (mic_data_size >> 3)] = abs(tmp_sum);
            // update new max peak
            if (correlate_data[channel][shift + (mic_data_size >> 3)] > peak_max[channel]) {
                peak_max[channel] = correlate_data[channel][shift + (mic_data_size >> 3)];
                peak_index[channel] = shift;
            }
        }
    }
}

// the thread waits for button push to record and call
// cross_correlate and compute_direction
static PT_THREAD(protothread_computation(struct pt * pt)) {
    PT_BEGIN(pt);
    while (1) {
        //wait till the push button is clicked
        while (!ready) {
            PT_YIELD(pt);
        }
        //reset the push button state
        ready = 0;
        //start DMA in all three channels
        DmaChnEnable(0);
        DmaChnEnable(2);
        DmaChnEnable(3);
        //wait until all DMA channel complete transferring entire block
        while (!(DmaChnGetEvFlags(0) & DMA_EV_DST_FULL &&
                DmaChnGetEvFlags(2) & DMA_EV_DST_FULL &&
                DmaChnGetEvFlags(3) & DMA_EV_DST_FULL))
            PT_YIELD(pt);
        //reset max-peak from previous computations
        int channel;
        for (channel = 0; channel < num_mic_channel; channel++)
            peak_max[channel] = 0;
        //cross-correlate data
        cross_correlate();
        //compute and display direction
        compute_direction();
    }
    PT_END(pt);
}

// This thread update and maintain the state of the push button
static PT_THREAD(protothread_button(struct pt * pt)) {
    PT_BEGIN(pt);
    while (1) {
        PT_YIELD_TIME_msec(30);
        //read digital input from B7
        button = mPORTBReadBits(BIT_7);
        // FSM for the button press debouncing
        switch (ButtonState) {
        case RELEASED:
            if (button) ButtonState = MAYBE_PRESSED;
            else ButtonState = RELEASED;
            break;
        case MAYBE_PRESSED:
            if (button) ButtonState = PRESSED;
            else ButtonState = MAYBE_PRESSED;
            break;
        case PRESSED:
            if (!button) ButtonState = MAYBE_RELEASED;
            else ButtonState = PRESSED;
            break;
        case MAYBE_RELEASED:
            if (!button) {
                ButtonState = RELEASED;
                // fully pressed, record new burst
                ready = 1;
            } else ButtonState = MAYBE_RELEASED;
            break;
        }
    } // END WHILE(1)
    PT_END(pt);
} // timer thread

//Debugging thread, to print useful debugging info to TFT display
static PT_THREAD(protothread_debug(struct pt * pt)) {
    PT_BEGIN(pt);
    while (1) {
        // yield time 0.5 second
        PT_YIELD_TIME_msec(500);
        sys_time_seconds++;
        //update display with useful debugging info      
        tft_fillRoundRect(10, 0, 200, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 0);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "Time=%d", sys_time_seconds);
        tft_writeString(buffer);
        tft_fillRoundRect(10, 20, 200, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 20);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "M0=%d", (short) ADC1BUF0);
        tft_writeString(buffer);
        tft_fillRoundRect(10, 40, 200, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 40);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "M1=%d", ADC1BUF1);
        tft_writeString(buffer);
        tft_fillRoundRect(10, 60, 200, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 60);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "M2=%d", ADC1BUF2);
        tft_writeString(buffer);
        tft_fillRoundRect(10, 80, 200, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 80);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "max=%d", peak_max[0]); //,peak_max[1],peak_max[2]);
        tft_writeString(buffer);
        tft_fillRoundRect(10, 100, 240, 16, 1, ILI9340_BLACK);
        tft_setCursor(10, 100);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(2);
        sprintf(buffer, "idx= %d %d %d", peak_index[0], peak_index[1], peak_index[2]);
        tft_writeString(buffer);
        // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
} // timer thread

// === Main  ======================================================
void main(void) {
    //SYSTEMConfigPerformance(PBCLK);

    ANSELA = 0;
    ANSELB = 0;
    /// SPI setup for DAC //////////////////////////////////////////
    // SCK2 is pin 26
    // SDO2 is in PPS output group 2, could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);
    // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_4);
    mPORTBSetBits(BIT_4);
    // divide Fpb by 2, configure the I/O ports. Not using SS in this example
    // 16 bit transfer CKP=1 CKE=1
    // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
    // For any given peripherial, you will need to match these
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV, spiClkDiv);

    // timer interrupt //////////////////////////
    // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
    // at 30 MHz PB clock 60 counts is two microsec
    // 400 is 100 ksamples/sec
    // 2000 is 20 ksamp/sec
    // 40000 is 1 ksamp/sec
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, sys_clock / sampling_rate);
    // set up the timer interrupt with a priority of 2
    //ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    //mT2ClearIntFlag(); // and clear the interrupt flag

    OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, sys_clock / 1000);
    ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
    mT3ClearIntFlag(); // and clear the interrupt flag

    // SCK2 is pin 26
    // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);
    // 16 bit transfer CKP=1 CKE=1
    // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
    // For any given peripherial, you will need to match these
    // NOTE!! IF you are using the port expander THEN
    // -- clk divider must be set to 4 for 10 MHz
    SpiChnOpen(SPI_CHANNEL2, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV, 4);
    // end DAC setup

    // the ADC ///////////////////////////////////////
    // configure and enable the ADC
    CloseADC10(); // ensure the ADC is off before setting the configuration

    // define setup parameters for OpenADC10
    // Turn module on | ouput in integer | trigger mode auto | enable autosample
    // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
    // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
    // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
    #
    define PARAM1 ADC_FORMAT_SIGN_INT16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON //

    // define setup parameters for OpenADC10
    // ADC ref external  | disable offset test | disable scan mode | do 3 sample | use single buf | alternate mode off
    # define PARAM2 ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_ON | ADC_SAMPLES_PER_INT_3 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
    //
    // Define setup parameters for OpenADC10
    // use peripherial bus clock | set sample time | set ADC clock divider
    // ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
    // ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
    # define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_15 | ADC_CONV_CLK_Tcy

    // define setup parameters for OpenADC10
    // set AN11 and  as analog inputs
    # define PARAM4 ENABLE_AN1_ANA | ENABLE_AN5_ANA | ENABLE_AN0_ANA //

    // define setup parameters for OpenADC10
    // DO not skip the channels you want to scan
    // do not specify channels 0, 1, 5
    // RA0 RA1 RB3 
    # define PARAM5 SKIP_SCAN_AN11 | SKIP_SCAN_AN2 | SKIP_SCAN_AN3 | SKIP_SCAN_AN4 | SKIP_SCAN_AN6 | SKIP_SCAN_AN7 | SKIP_SCAN_AN8 | SKIP_SCAN_AN9 | SKIP_SCAN_AN10 | SKIP_SCAN_AN12 | SKIP_SCAN_AN13 | SKIP_SCAN_AN14 | SKIP_SCAN_AN15

    // use ground as neg ref for A
    // actual channel number is specified by the scan list
    SetChanADC10(ADC_CH0_NEG_SAMPLEA_NVREF); //
    OpenADC10(PARAM1, PARAM2, PARAM3, PARAM4, PARAM5); // configure ADC using the parameters defined above

    EnableADC10(); // Enable the ADC
    ///////////////////////////////////////////////////////

    // DMA channel 0
    DmaChnOpen(0, 2, DMA_OPEN_DEFAULT);
    // channel, channel priority, mode
    DmaChnSetTxfer(0, (void * ) & ADC1BUF0, (void * ) mic_data[0], 2, mic_data_size * 2, 2);
    // (ch, start virtual addr, dest virtual addr, source size,
    // dest size, cell size)
    // 256 16-bit integers
    DmaChnSetEventControl(0, DMA_EV_START_IRQ(_TIMER_2_IRQ));
    DmaChnWriteEvEnableFlags(0, DMA_EV_DST_FULL);
    // channel, trigger IRQ - Trigger on ADC conversion done
    //DmaChnEnable(0);

    // DMA channel 2
    DmaChnOpen(2, 2, DMA_OPEN_DEFAULT);
    // channel, channel priority, mode
    DmaChnSetTxfer(2, (void * ) & ADC1BUF1, (void * ) mic_data[1], 2, mic_data_size * 2, 2);
    // (ch, start virtual addr, dest virtual addr, source size,
    // dest size, cell size)
    // 256 16-bit integers
    DmaChnSetEventControl(2, DMA_EV_START_IRQ(_TIMER_2_IRQ));
    DmaChnWriteEvEnableFlags(2, DMA_EV_DST_FULL);
    // channel, trigger IRQ - Trigger on ADC conversion done
    //DmaChnEnable(2);

    // DMA channel 3
    DmaChnOpen(3, 2, DMA_OPEN_DEFAULT);
    // channel, channel priority, mode
    DmaChnSetTxfer(3, (void * ) & ADC1BUF2, (void * ) mic_data[2], 2, mic_data_size * 2, 2);
    // (ch, start virtual addr, dest virtual addr, source size,
    // dest size, cell size)
    // 256 16-bit integers
    DmaChnSetEventControl(3, DMA_EV_START_IRQ(_TIMER_2_IRQ));
    DmaChnWriteEvEnableFlags(3, DMA_EV_DST_FULL);
    // channel, trigger IRQ - Trigger on ADC conversion done
    //DmaChnEnable(3);

    // === set up i/o port pin ===============================
    //mPORTBSetPinsDigitalOut(BIT_0 );  //Set port as output
    // PortB as inputs
    mPORTBSetPinsDigitalIn(BIT_7); //Set port as input
    // and turn on pull-down on inputs
    EnablePullDownB(BIT_7);

    // === config threads ==========
    // turns OFF UART support and debugger pin, unless defines are set
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();

    // init the threads
    PT_INIT( & pt_debug);
    PT_INIT( & pt_button);
    PT_INIT( & pt_computation);

    // init the display
    // NOTE that this init assumes SPI channel 1 connections
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    //240x320 vertical display
    tft_setRotation(0); // Use tft_setRotation(1) for 320x240

    // seed random color
    srand(1);

    // round-robin scheduler for threads
    while (1) {
        PT_SCHEDULE(protothread_debug( & pt_debug));
        PT_SCHEDULE(protothread_button( & pt_button));
        PT_SCHEDULE(protothread_computation( & pt_computation));
    }
} // main

// === end  ======================================================